import sys

import click


from .config import resolve_config, parse_config


@click.group()
def main() -> None:
    pass


@main.command()
@click.argument("ENV_FILE", type=str, required=True)
def shell(env_file: str) -> None:
    import shlex

    for k, v in parse_config(resolve_config(env_file)).items():
        click.echo(f"{k}={shlex.quote(v)}")


@main.command()
@click.option(
    "-a",
    "--at",
    type=str,
    prompt="You can provide a date, an ISO-like timestamp, or something like '10 days ago'\nFrom when?",
)
@click.option(
    "-c", "--config", type=str, required=True, help="path or name of config file"
)
@click.argument("EXTRACT_FILE", type=str, required=True)
@click.argument("OUTPUT_FILE", type=click.Path(allow_dash=True))
def extract_file_at(at: str, config: str, extract_file: str, output_file: str) -> None:
    """
    Extract a file from the git history from a particular datetime

    \b
    EXTRACT_FILE is the name of the file, e.g. todo.txt
    OUTPUT_FILE is the file to write to, or '-' to write to STDOUT
    """
    import dateparser

    from . import DocHistory, naive_dt

    conf = DocHistory.from_dict(parse_config(resolve_config(config)))
    dt = dateparser.parse(at)
    if dt is None:
        click.echo(f"Couldnt parse {at} into a datetime", err=True)
        sys.exit(1)

    data_bytes = conf.extract_buffer_at(extract_file, naive_dt(dt))
    if output_file == "-":
        click.echo(data_bytes.decode("utf-8"), nl=False)
    else:
        with open(output_file, "wb") as f:
            f.write(data_bytes)


if __name__ == "__main__":
    main(prog_name="git_doc_history")



import sys
import os
from pathlib import Path
from typing import Dict, Union

import click
from dotenv import dotenv_values

config_dir = os.environ.get(
    "XDG_CONFIG_DIR", os.path.join(os.environ["HOME"], ".config")
)


def _resolve_env_file(conf: str) -> str:
    if os.path.exists(conf):
        return conf
    dconf = Path(config_dir) / "git_doc_history" / conf
    if dconf.exists():
        return str(dconf)
    raise FileNotFoundError(f"Could not find config file at {conf} or {dconf}")


def resolve_config(path: str) -> Path:
    try:
        conf = Path(_resolve_env_file(path)).absolute()
    except FileNotFoundError as fe:
        click.echo(str(fe), err=True)
        sys.exit(1)
    return conf


def parse_config(input_file: Union[Path, str]) -> Dict[str, str]:
    res: Dict[str, str] = {}
    for k, v in dotenv_values(Path(input_file)).items():
        assert v is not None
        if os.linesep not in v and k != "COPY_FILES":
            res[k] = str(Path(v).expanduser().absolute())  # expand directory names
        else:
            res[k] = v  # dont modify lines with multiple lines
    assert "SOURCE_DIR" in res, "Missing SOURCE_FILES key"
    assert "BACKUP_DIR" in res, "Missing BACKUP_DIR key"
    assert "COPY_FILES" in res, "Missing COPY_FILES key"
    return res



import io
from pathlib import Path
from typing import (
    NamedTuple,
    Literal,
    List,
    Dict,
    Set,
    Any,
    Optional,
    Iterator,
    Callable,
)
from datetime import datetime

from git.repo import Repo
from git.objects import Commit


def naive_dt(dt: datetime) -> datetime:
    return datetime.fromtimestamp(dt.timestamp())


def commit_local_time(commit: Commit) -> datetime:
    return naive_dt(commit.committed_datetime)


class DocHistorySnapshot(NamedTuple):
    commit_sha: str
    epoch_time: int
    data: bytes

    @property
    def dt(self) -> datetime:
        return datetime.fromtimestamp(self.epoch_time)


class DocHistory(NamedTuple):
    """
    Traverses and/or extracts files from git history
    """

    backup_dir: Path
    copy_files: List[str]
    source_dir: Optional[Path] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "DocHistory":
        return cls(
            source_dir=Path(data["SOURCE_DIR"]).expanduser().absolute(),
            backup_dir=Path(data["BACKUP_DIR"]).expanduser().absolute(),
            copy_files=data["COPY_FILES"].strip().splitlines(),
        )

    def repo(self) -> Repo:
        return Repo(str(self.backup_dir))

    def extract_bytes_from_commit(self, commit: Commit, file: str) -> bytes:
        try:
            blob = commit.tree / file
        except KeyError as k:
            #  assert file in self.copy_files, f"{file} not in {self.copy_files}"
            raise k
        with io.BytesIO(blob.data_stream.read()) as f:
            data = f.read()
        return data

    def extract_str_from_commit(self, commit: Commit, file: str) -> str:
        return self.extract_bytes_from_commit(commit, file).decode("utf-8")

    def iter_commit_snapshots(self, file: str) -> Iterator[DocHistorySnapshot]:
        for commit in self.repo().iter_commits():
            try:
                blob = self.extract_bytes_from_commit(commit, file)
            except (KeyError, ValueError):
                pass  # file doesn't exist at this hash
            else:
                yield DocHistorySnapshot(
                    commit_sha=commit.hexsha,
                    epoch_time=commit.committed_date,
                    data=blob,
                )

    def list_commit_snapshots(
        self, file: str, reverse: bool = False
    ) -> List[DocHistorySnapshot]:
        items = list(self.iter_commit_snapshots(file))
        if reverse:
            items = list(reversed(items))
        return items

    def extract_buffer_at(self, file: str, at: datetime) -> bytes:
        """
        Given a file in backup_dir, fetch the contents at datetime
        """
        assert (
            at.tzinfo is None
        ), "Datetime must be naive -- to compare with sorted commits"
        # default to first commit -- latest
        commit: Optional[Commit] = None
        for comm in self.repo().iter_commits():
            if commit is None:
                commit = comm
            if commit_local_time(comm) < at:
                break
            commit = comm
        if commit is None:
            raise RuntimeError("No commit set -- likely no commits in repo?")
        return self.extract_bytes_from_commit(commit, file)


Action = Literal["added", "removed"]


class Diff(NamedTuple):
    """
    A diff -- whats changed from one git snapshot to the next
    """

    epoch_time: int
    data: Any
    action: Action

    @property
    def description(self) -> str:
        return f"{self.action} {datetime.fromtimestamp(self.epoch_time)} {self.data}"


def parse_lines(snapshot: DocHistorySnapshot) -> List[str]:
    """
    The most basic parse_func -- decodes to text and returns a list of lines
    """
    return snapshot.data.decode("utf-8").strip().splitlines()


def parse_snapshot_diffs(
    doc: "DocHistory",
    file: str,
    parse_func: Optional[Callable[[DocHistorySnapshot], List[Any]]] = None,
) -> Iterator[Diff]:
    """
    Iterates through the git history in chronological order, keeping track
    of when data was added or removed. By default, this parses the 'file'
    given by splitting it into lines. If lines are added/removed, this returns an
    event which specifies when in the history, and what was added/removed

    Alternatively, can pass a 'parse_func', which is a function which
    accepts the DocHistorySnapshot, and retuns a list of hashable items
    to store as state
    """
    if parse_func is None:
        parse_func = parse_lines
    data = doc.list_commit_snapshots(file=file, reverse=True)
    state: Set[Any] = set()

    for snapshot in data:
        # parse the data using the func
        snapshot_data: List[Any] = parse_func(snapshot)
        snapshot_set = set()
        # e.g. iterate over lines -- or whatever data was parsed
        for sn in snapshot_data:

            snapshot_set.add(sn)
            # we're already tracking this in state -- it was added in a previous commit
            if sn in state:
                continue
            state.add(sn)
            yield Diff(epoch_time=snapshot.epoch_time, data=sn, action="added")

        # something is removed, if its in state, but its not in
        # the currently passed data
        for sn in list(state):
            if sn not in snapshot_set:
                yield Diff(epoch_time=snapshot.epoch_time, data=sn, action="removed")
                state.remove(sn)
